! WHIZARD 1.23 Dec 10 2002
! 
! (C) 1999-2002 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module cascades

  use kinds, only: default !NODEP!
  use file_utils !NODEP!
  use limits, only: BUFFER_SIZE, MAX_CASCADES, FILENAME_LEN
  use files, only: file_exists_else_default
  use parser
  use particles
  use events, only: write_process
  use decay_trees, only: TC, decay_level
  use diagnostics, only: msg_buffer, msg_message, msg_error, msg_fatal
  use parameters, only: parameter_set !NODEP!

  implicit none
  private

  type, public :: vertex
     integer, dimension(3) :: code
  end type vertex

  type, private :: node_t 
     integer(kind=TC) :: bincode
     integer :: code, index
     real(kind=default) :: mass
     logical :: resonant, decays, t_channel, initial, final, needs_cut
     type(node_t), pointer :: daughter1, daughter2
  end type node_t

  type, public :: cascade
     private
     logical :: active, subdominant, deleted, final
     integer :: depth, multiplicity, n_t_channel, n_resonances
     integer :: grove
     integer(kind=TC), dimension(:), pointer :: tree
     integer, dimension(:), pointer :: code
     logical, dimension(:), pointer :: resonant, t_channel
     type(node_t), pointer :: node
  end type cascade

  type, public :: phase_space_switches
     integer :: off_shell_lines
     integer :: exchange_lines
     integer :: extra_off_shell_lines
     logical :: s_channel_resonance
     logical :: single_off_shell_decays
     logical :: double_off_shell_decays
     logical :: single_off_shell_branchings
     logical :: double_off_shell_branchings
     logical :: massive_fsr
     logical :: massless_splittings
     integer :: splitting_depth
     real(kind=default) :: splitting_scale
     real(kind=default) :: threshold_mass
  end type phase_space_switches

  type(vertex), parameter, public :: NO_VERTEX = vertex((/0,0,0/))

  public :: read
  public :: write
  public :: scan
  public :: setup_and_read_vertices
  public :: clear_vertices
  public :: write_vertices
  public :: test_vertices
  public :: create, destroy
  public :: cascade_set_write
  public :: cascade_set_init
  public :: cascade_set_clear
  public :: cascade_set_add
  public :: cascade_set_generate
  public :: cascade_set_select
  public :: cascade_set_validate
  public :: cascade_set_groves
  public :: make_phase_space_file

  type(vertex), dimension(:), allocatable, save :: vertex_set
  type(cascade), dimension(MAX_CASCADES), save :: cascade_set
  integer, save :: n_cascades, n_groves
  integer, dimension(MAX_CASCADES), save :: cascade_hash
  logical, save :: single_off_shell_decays
  logical, save :: double_off_shell_decays
  logical, save :: single_off_shell_branchings
  logical, save :: double_off_shell_branchings
  logical, save :: massive_FSR
  logical, save :: massless_splittings
  integer, save :: splitting_depth

  interface read
     module procedure vertex_read_unit
  end interface
  interface write
     module procedure vertex_write_unit
  end interface
  interface scan
     module procedure vertices_scan_file
  end interface
  interface read
     module procedure vertices_read_file
  end interface
  interface write_vertices
     module procedure write_vertices_unit
  end interface
  interface create
     module procedure cascade_create
  end interface
  interface clear
     module procedure cascade_clear
  end interface
  interface destroy
     module procedure cascade_destroy
  end interface

  interface write
     module procedure cascade_write_unit
  end interface
  interface cascade_set_write
     module procedure cascade_set_write_unit
  end interface
  interface cascade_set_generate
     module procedure cascade_set_generate_s
     module procedure cascade_set_generate_t
  end interface

contains

  subroutine vertex_read_unit (u, v, ok)
    integer, intent(in) :: u
    type(vertex), intent(out) :: v
    logical, intent(out), optional :: ok
    character(len=BUFFER_SIZE) :: string
    integer :: i
    if (present(ok)) ok = .true.
    v = NO_VERTEX
    call find_string (u, "vertex", ok=ok)
    if (present(ok)) then
       if (.not.ok) return
    end if
    do i=1,3
       string = get_string (u)
       v%code(i) = particle_code (trim(string))
    end do
  end subroutine vertex_read_unit

  subroutine vertex_write_unit (u, v)
    integer, intent(in) :: u
    type(vertex), intent(in) :: v
    write(u, '(4(1x,A))') "vertex    ", &
         & particle_name(v%code(1)), &
         & particle_name(v%code(2)), &
         & particle_name(v%code(3))
  end subroutine vertex_write_unit
     
  subroutine vertex_check (v, ok)
    type(vertex), intent(in) :: v
    logical, intent(out) :: ok
    integer, dimension(3) ::  multiplicity, charge
    integer :: i
    do i=1,3
       multiplicity(i) = particle_spin_type (v%code(i))-1
       charge(i) = particle_3charge (v%code(i))
    end do
    if (any((v%code)==UNDEFINED)) then
       call write (6, v)
       call msg_error (" This vertex contains an unknown particle.  I'll skip it.")
       ok = .false.
    else if (mod(sum(multiplicity),2)/=0) then
       call write (6, v)
       call msg_error (" This vertex has odd fermion number.  I'll skip it.")
       ok = .false.
    else if (sum(charge)/=0) then
       call write (6, v)
       call msg_error (" This vertex has nonzero charge.  I'll skip it.")
       ok = .false.
    else
       ok = .true.
    end if
  end subroutine vertex_check

  subroutine vertices_scan_file (file, count)
    character(len=*), intent(in) :: file
    integer, intent(out), optional :: count
    integer :: u
    type(vertex) :: v
    logical :: ok
    u = free_unit()
    open (unit=u, file=file, status='old', action='read')
    ok = .true.
    if (present(count)) then
       Count = 0
    else
       call msg_message (" Vertices in "//trim(file)//":")
    end if
    SCAN: do
       call read (u, v, ok)
       if (.not.ok) exit SCAN
       call vertex_check (v, ok)
       if (.not.ok) cycle SCAN
       if (present(count)) then
          count = count + 1
       else
          call write (6, v)
       end if
    end do SCAN
    close (u)
  end subroutine vertices_scan_file

  subroutine vertices_read_file (file, vv)
    character(len=*), intent(in) :: file
    type(vertex), dimension(:), intent(inout) :: vv
    integer :: u, i
    u = free_unit()
    open (unit=u, file=file, status='old', action='read')
    do i=1, size(vv)
       call read (u, vv(i))
    end do
    close (u)
  end subroutine vertices_read_file

  subroutine setup_and_read_vertices (prefix, filename)
    character(len=*), intent(in) :: prefix, filename
    character(len=FILENAME_LEN) :: file
    integer :: count
    file = file_exists_else_default (prefix, filename, "mdl")
    if (len_trim (file) == 0) call msg_fatal &
         & (" Model configuration file not found.")
    call msg_message (" Reading vertices from file "//trim(file)//" ...")
    call scan (trim(file), count)
    if (count==0)  call msg_fatal (" Model file: No vertices found.")
    write (msg_buffer, "(1x,A,2x,I5,1x,A)") &
         & "Model file:", count, "vertices found."
    call msg_message
    allocate (vertex_set(count))
    call read (trim(file), vertex_set)
  end subroutine setup_and_read_vertices

  subroutine clear_vertices
    deallocate (vertex_set)
  end subroutine clear_vertices
  subroutine write_vertices_unit (u)
    integer, intent(in) :: u
    integer :: i
    do i=1, size(vertex_set)
       call write (u, vertex_set(i))
    end do
  end subroutine write_vertices_unit

  subroutine lookup_vertex (code1, code2, code3, point)
    integer, intent(in) :: code1, code2
    integer, intent(inout) :: code3
    integer, intent(inout) :: point
    integer :: code
    code = UNDEFINED
    SCAN: do
       if (point > size(vertex_set)) then
          code3 = UNDEFINED
          exit SCAN
       end if
       if (match(1,2).or.match(2,1)) then
          code = vertex_set(point)%code(3)
       else if (match(1,3).or.match(3,1)) then
          code = vertex_set(point)%code(2)
       else if (match(2,3).or.match(3,2)) then
          code = vertex_set(point)%code(1)
       end if
       point = point+1
       if (code /= UNDEFINED) then
          if (code3 == UNDEFINED) then
             code3 = code
             exit SCAN
          else if (code == code3) then
             exit SCAN
          else
             code = UNDEFINED
          end if
       end if
    end do SCAN
  contains
    function match(i,j)
      integer, intent(in) :: i,j
      logical :: match
      match = (code1 == vertex_set(point)%code(i)) &
           & .and. (code2 == vertex_set(point)%code(j))
    end function match
  end subroutine lookup_vertex

  subroutine test_vertices (code1, code2)
    integer, intent(in) :: code1, code2
    integer :: point, code3
    point = 0
    write(*,'(3(1x,A),A)') "Particles matching ", &
         & trim(particle_name(code1)), trim(particle_name(code2)), ":"
    SCAN: do
       code3 = UNDEFINED
       call lookup_vertex (code1, code2, code3, point)
       if (code3/=UNDEFINED) then
          print *, particle_name(code3)
       else
          exit SCAN
       end if
    end do SCAN
  end subroutine test_vertices

  subroutine node_create (node)
    type(node_t), pointer :: node
    node%bincode = 0
    node%code = UNDEFINED
    node%index = 0
    node%mass = 0
    node%resonant = .false.
    node%decays = .false.
    node%t_channel = .false.
    node%initial = .false.
    node%final = .false.
    node%needs_cut = .false.
    nullify(node%daughter1)
    nullify(node%daughter2)
  end subroutine node_create

  subroutine cascade_create (cs, depth)
    type(cascade), intent(inout) :: cs
    integer, intent(in) :: depth
    allocate(cs%node)
    allocate(cs%tree(depth))
    allocate(cs%code(depth))
    allocate(cs%resonant(depth))
    allocate(cs%t_channel(depth))
    call clear (cs)
  end subroutine cascade_create

  subroutine cascade_clear (cs)
    type(cascade), intent(inout) :: cs
    cs%active = .false.
    cs%subdominant = .false.
    cs%deleted = .false.
    cs%final = .false.
    cs%depth = 0
    cs%multiplicity = 0
    cs%n_t_channel = -1
    cs%n_resonances = 0
    cs%grove = 0
    cs%tree = 0
    cs%code = UNDEFINED
    cs%resonant = .false.
    cs%t_channel = .false.
    call node_create (cs%node)
  end subroutine cascade_clear

  subroutine cascade_destroy (cs)
    type(cascade), intent(inout) :: cs
    deallocate(cs%node)
    deallocate(cs%tree)
    deallocate(cs%code)
    deallocate(cs%resonant)
    deallocate(cs%t_channel)
  end subroutine cascade_destroy

  subroutine cascade_write_unit (u, cs, final, short, init)
    integer, intent(in) :: u
    type(cascade), intent(in) :: cs
    logical, intent(in), optional :: final, short, init
    logical :: short_form, final_only
    integer(kind=TC), dimension(cs%depth) :: tree
    integer :: i, d, count
    if (present(init)) then 
       count = 0
       return
    end if
1   format(1x,A,11(1x,I3))
2   format(2x,A,1x,I3,1x,A,1x,A)
    short_form = .false.;  if (present(short)) short_form = short
    final_only = .false.;  if (present(final)) final_only = final
    if (cs%active) then
       if (short_form) then
          if (cs%deleted) then
             call start_comment(u)
          else
             count = count + 1
          end if
          tree = 0
          if (final_only) then
             d = (cs%depth-3)/2
             tree = reduced(cs%tree(:cs%depth-1))
          else
             d = cs%depth
             tree = cs%tree
          end if
          write(u,1,advance='no') 'tree', tree(:d)
          write(u,'(5x)',advance='no')
          call start_comment(u)
          if (cs%deleted) then
             write (u, '(6x)', advance='no')
          else
             write(u,'(1x,I4,A,1x)', advance='no') count, "*"
          end if
          do i=1, cs%depth
             if (any(tree(:d)==cs%tree(i))) write(u,'(1x,A)',advance='no') &
                  & trim(particle_name(cs%code(i)))
          end do
          write(u,*)
          do i=1, cs%depth
             if (cs%resonant(i)) then
                if (cs%deleted) call start_comment (u)
                if (cs%t_channel(i)) then
                   if (cs%node%daughter1%t_channel) then
                      write(u,2) ' map', cs%node%daughter1%bincode, &
                           & 't-channel ', &
                           & trim(particle_name(cs%node%daughter1%code))
                   else
                      write(u,2) ' map', cs%node%daughter2%bincode, &
                           & 't-channel ', &
                           & trim(particle_name(cs%node%daughter2%code))
                   end if
                else
                   write(u,2) ' map', cs%tree(i), 's-channel ', &
                        & trim(particle_name(cs%code(i)))
                end if
             end if
          end do
       else
          write(u,*) 'Cascade #', cs%node%index
          write(u,*) '  Grove:      #', cs%grove
          write(u,*) '  Depth:       ', cs%depth
          write(u,*) '  Multiplicity:', cs%multiplicity
          write(u,*) '  Resonances:  ', cs%n_resonances
          write(u,*) '  T-channel:   ', cs%n_t_channel
          if (final_only) then
             write(u,*) '  Tree:      ', reduced(cs%tree(:cs%depth-1))
          else
             write(u,*) '  Tree:      ', cs%tree(:cs%depth)
          end if
          if (cs%node%needs_cut) &
               & write(u,*) '  Zero-mass singularity'
          call node_write_unit (u, cs%node)
       end if
    else if (.not.short_form) then
       write(u,*) 'Cascade #', cs%node%index, ': deleted'
    end if
  contains
    function reduced (array)
      integer(kind=TC), dimension(:), intent(in) :: array
      integer(kind=TC), dimension((size(array))/2-1) :: reduced
      integer :: i, j
      j = 1
      do i=1, size(array)
         if (decay_level(array(i)) > 1) then
            reduced(j) = array(i)
            j = j+1
         end if
      end do
    end function reduced
  end subroutine cascade_write_unit

  recursive subroutine node_write_unit (u, node)
    integer, intent(in) :: u
    type(node_t), pointer :: node, daughter1, daughter2
    call entry_write_unit (u, node)
    if (node%decays) then
       daughter1 => node%daughter1
       daughter2 => node%daughter2
       write(u, '(1x,A)', advance='no') '->'
       call entry_write_unit (u, daughter1)
       call entry_write_unit (u, daughter2)
       write(u,*)
       call node_write_unit (u, daughter1)
       call node_write_unit (u, daughter2)
    else
       write(u,*)
    end if
  contains
    subroutine entry_write_unit (u, node)
      integer, intent(in) :: u
      type(node_t), pointer :: node
      write(u, 1, advance='no') &
           & trim(particle_name(node%code)), '(', node%index, &
           & ',', node%mass, ')'
      if (node%resonant) then
         write(u, '(A)', advance='no')  "*"
      end if
      if (node%initial .or. node%final) then
         write(u, '(A)', advance='no') "+"
      else if (.not.node%decays) then
         write(u, '(A)', advance='no') "x"
      else if (node%t_channel) then
         write(u, '(A)', advance='no') "t"
      end if
1     format(5x,A,A,I4,A,1x,1PE8.2,A)
    end subroutine entry_write_unit
  end subroutine node_write_unit

  subroutine start_comment (u)
    integer, intent(in) :: u
    write(u, '(1x,A)', advance='no') '!'
  end subroutine start_comment

  subroutine cascade_set_write_unit (u, final, short, sort)
    integer, intent(in) :: u
    logical, intent(in), optional :: final, short, sort
    integer :: i, grove, n_r, n_t
    logical :: sorted, final_only, short_form
    final_only = .false.;  if (present(final)) final_only = final
    short_form = .false.;  if (present(short)) short_form = short
    sorted = .false.;  if (present(sort)) sorted = sort
    call write (u, cascade_set(1), init=.true.)
    SCAN_GROVES: do grove=0, n_groves
       if (short_form .and. sorted .and. &
            & .not.(grove==0 .and. final_only)) then
          FIND_FIRST_MATCH: do i=1, n_cascades
             if (cascade_set(i)%grove==grove) then
                n_r = cascade_set(i)%n_resonances
                n_t = cascade_set(i)%n_t_channel
                if (cascade_set(i)%deleted) call start_comment (u)
                call start_comment (u)
                write(u,'(1x,A,I2,A)', advance='no') 'Multiplicity =', &
                     & cascade_set(i)%multiplicity, ","
                if (n_r==0) then
                   write(u,'(1x,A)', advance='no') 'no resonances, '
                else if (n_r==1) then
                   write(u,'(1x,A)', advance='no') ' 1 resonance,  '
                else
                   write(u,'(1x,I2,1x,A)', advance='no') n_r, 'resonances, '
                end if
                if (n_t==0) then
                   write(u,'(1x,A)') 's-channel graph'
                else if (n_t==1) then
                   write(u,'(1x,A)') ' 1 t-channel line'
                else
                   write(u,'(1x,I2,1x,A)') n_t, 't-channel lines'
                end if
                if (cascade_set(i)%deleted) call start_comment (u)
                write(u,'(1x,A,I3)') 'grove #', grove
                exit FIND_FIRST_MATCH
             end if
          end do FIND_FIRST_MATCH
       end if
       do i=1, n_cascades
          if ((cascade_set(i)%grove==grove .or. .not.sorted) .and. &
               & (cascade_set(i)%final .or..not.final_only)) then
             call write (u, cascade_set(i), final, short)
          end if
       end do
       if (.not.sorted) exit SCAN_GROVES
       write(u,*)
    end do SCAN_GROVES
    if (present(final)) then
       if (final) then
          write (u,*) "! Number of channels total:   ", &
               & count (cascade_set%active .and. cascade_set%final)
          write (u,*) "! Number of channels kept:    ", &
               & count (cascade_set%active .and. cascade_set%final .and. &
               &        .not. cascade_set%deleted)
          write (u,*) "! Number of dominant channels:", &
               & count (cascade_set%active .and. cascade_set%final .and. &
               &        .not. cascade_set%subdominant)
       else
          write (u,*) "! Number of cascades shown:", &
               & count (cascade_set%active)
       end if
    end if
  end subroutine cascade_set_write_unit

  subroutine cascade_set_init (code, par, threshold_mass)
    integer, dimension(:,:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    integer :: i, f, n_out, n_flv
    n_cascades = 0
    n_groves = 0
    call cascade_set_clear
    n_out = size(code, dim=1)
    n_flv = size(code, dim=2)
    do i=1, n_out
       do f=1, n_flv
          if (all (code(i,1:f-1) /= code(i,f))) then
             call cascade_set_add &
                  & (ibset(0_TC,i-1), code(i,f), par, threshold_mass, &
                  &  initial=.false.)
          end if
       end do
    end do
    cascade_hash = 0
  end subroutine cascade_set_init

  subroutine cascade_set_clear
    integer :: i
    cascade_set%active = .false.
    cascade_set%subdominant = .false.
    cascade_set%deleted = .false.
    cascade_set%final = .false.
    cascade_set%depth = 0
    cascade_set%multiplicity = 0
    cascade_set%n_t_channel = -1
    cascade_set%n_resonances = 0
    cascade_set%grove = 0
    do i=1, n_cascades
       call destroy (cascade_set(i))
    end do
    n_cascades = 0
    n_groves = 0
  end subroutine cascade_set_clear

  subroutine cascade_set_add &
       & (bincode, code, par, threshold_mass, initial, index)
    integer(kind=TC), intent(in) :: bincode
    integer, intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    logical, intent(in) :: initial
    integer, intent(out), optional :: index
    integer :: k
    k = n_cascades+1
    if (k>size(cascade_set)) then
       write (msg_buffer, "(1x,A,1x,I6,1x,A)") &
            & "More than ", MAX_CASCADES, &
            & "cascades.  You may increase MAX_CASCADES in limits.f90"
       call msg_fatal
    end if
    call cascade_create (cascade_set(k), 1)
    cascade_set(k)%active = .true.
    cascade_set(k)%depth = 1
    cascade_set(k)%node%bincode = bincode
    cascade_set(k)%node%code = code
    cascade_set(k)%node%index = k
    if (particle_mass (code, par) > threshold_mass) then
       cascade_set(k)%node%mass = particle_mass (code, par)
    else
       cascade_set(k)%node%mass = 0
    end if
    cascade_set(k)%node%resonant = .false.
    cascade_set(k)%node%initial = initial
    cascade_set(k)%node%t_channel = initial
    if (initial) then
       cascade_set(k)%multiplicity = 0
    else
       cascade_set(k)%multiplicity = 1
    end if
    cascade_set(k)%tree(1) = cascade_set(k)%node%bincode
    cascade_set(k)%code(1) = cascade_set(k)%node%code
    cascade_set(k)%t_channel(1) = initial
    if (initial) cascade_set(k)%n_t_channel = 0
    if (present(index)) index = k
    n_cascades = n_cascades + 1
  end subroutine cascade_set_add

  subroutine cascade_set_generate_s (par, threshold_mass, off_shell)
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    integer, intent(in), optional :: off_shell
    integer :: i, j
    i = 0
    OUTER_LOOP: do
       i = i+1;  if (i>n_cascades) exit OUTER_LOOP
       j = 0
       INNER_LOOP: do
          j = j+1;  if (j>n_cascades) exit INNER_LOOP
          call cascade_match &
               & (i, j, decay=.true., &
               &  par=par, threshold_mass=threshold_mass, off_shell=off_shell)
       end do INNER_LOOP
    end do OUTER_LOOP
  end subroutine cascade_set_generate_s

  subroutine cascade_set_generate_t &
       & (par, threshold_mass, depth, seed, target, off_shell)
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    integer, intent(in) :: depth, seed, target
    integer, intent(in), optional :: off_shell
    integer :: i, j
    i = 0
    OUTER_LOOP: do
       i = i+1;  if (i>n_cascades) exit OUTER_LOOP
       if (.not.cascade_disjunct(i,target)) cycle OUTER_LOOP
       j = 0
       INNER_LOOP: do
          j = j+1;  if (j>n_cascades) exit INNER_LOOP
          if (.not.cascade_disjunct(j,target)) cycle INNER_LOOP
          if (cascade_disjunct(i,seed) .and. cascade_disjunct(j,seed)) &
               & cycle INNER_LOOP
          call cascade_match &
               & (i, j, decay=.false., &
               &  par=par, threshold_mass=threshold_mass, &
               &  depth=depth, target=target, off_shell=off_shell)
       end do INNER_LOOP
    end do OUTER_LOOP
  end subroutine cascade_set_generate_t

  function cascade_disjunct (i, j) result (disjunct)
    integer, intent(in) :: i, j
    logical :: disjunct
    disjunct = (iand (cascade_set(i)%node%bincode, &
         &            cascade_set(j)%node%bincode) == 0)
  end function cascade_disjunct

  subroutine cascade_match &
       & (i, j, decay, par, threshold_mass, depth, target, off_shell)
    integer, intent(in) :: i, j
    logical, intent(in) :: decay
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    integer, intent(in), optional :: depth, target, off_shell
    integer :: point, d1, d2, d, code1, code2, code3
    logical :: ok, resonant, final, needs_cut
    real(kind=default) :: mass, mass_sum
    d1 = cascade_set(i)%depth
    d2 = cascade_set(j)%depth
    d = d1 + d2 + 1
    if (present(target).and.present(depth)) then
       if (d>depth) then
          return
       else
          final = (d == depth)
       end if
    else
       final = .false.
    end if
    if (cascade_disjunct(i,j)) then
       point = 1
       code1 = cascade_set(i)%node%code
       code2 = cascade_set(j)%node%code
       if (final) code3 = cascade_set(target)%node%code
       SCAN_VERTICES: do
          if (.not.final)  code3 = UNDEFINED
          call lookup_vertex (code1, code2, code3, point)
          if (code3==UNDEFINED) exit SCAN_VERTICES
          if (decay) then
             call cascade_check_decay &
                  & ((/i,j/), code3, par, threshold_mass, &
                  &  ok, resonant, needs_cut, mass, mass_sum)
          else 
             if (cascade_set(i)%node%t_channel) then
                call cascade_check_exchange &
                     & ((/i,j/), code3, par, threshold_mass, final, &
                     &  ok, resonant, needs_cut, mass)
             else
                call cascade_check_exchange &
                     & ((/j,i/), code3, par, threshold_mass, final, &
                     &  ok, resonant, needs_cut, mass)
             end if
             mass_sum = mass
          end if
          if (ok)  call cascade_merge &
               & (i, j, antiparticle_code(code3), &
               &  resonant, needs_cut, mass, final, off_shell)
          if (ok.and.resonant)  call cascade_merge &
               & (i, j, antiparticle_code(code3), &
               &  .false., needs_cut, mass_sum, final, off_shell)
       end do SCAN_VERTICES
    end if
  end subroutine cascade_match

  subroutine cascade_merge &
       & (i, j, code, resonant, needs_cut, mass, final, off_shell)
    integer, intent(in) :: i, j, code
    logical, intent(in) :: resonant, needs_cut, final
    integer, intent(in), optional :: off_shell
    real(kind=default), intent(in) :: mass
    integer :: k, d, d1, d2, n_t_channel, n_resonances, hash, internal
    integer(kind=TC) :: bc1, bc2, bc
    integer(kind=TC), dimension(:), allocatable :: tree
    integer, dimension(:), allocatable :: codes
    logical, dimension(:), allocatable :: reson, tchan
    logical :: ok
    ok = .true.
    d1 = cascade_set(i)%depth
    d2 = cascade_set(j)%depth
    d = d1 + d2 + 1
    internal = (d-3)/2
    n_resonances = cascade_set(i)%n_resonances + cascade_set(j)%n_resonances
    if (resonant) n_resonances = n_resonances + 1
    if (present(off_shell)) then
       if (internal - n_resonances > off_shell)  return
    end if
    bc1 = cascade_set(i)%node%bincode
    bc2 = cascade_set(j)%node%bincode
    bc = ior (bc1, bc2)
    allocate(tree(d), codes(d), reson(d), tchan(d))
    tree(1:d1)  = cascade_set(i)%tree
    codes(1:d1) = cascade_set(i)%code
    reson(1:d1) = cascade_set(i)%resonant
    tchan(1:d1) = cascade_set(i)%t_channel
    tree(d1+1:d1+d2)  = cascade_set(j)%tree
    codes(d1+1:d1+d2) = cascade_set(j)%code
    reson(d1+1:d1+d2) = cascade_set(j)%resonant
    tchan(d1+1:d1+d2) = cascade_set(j)%t_channel
    tree(d)  = bc
    codes(d) = code
    reson(d) = resonant
    tchan(d) = any(tchan(:d-1))
    call tree_sort (tree, codes, reson, tchan)
    hash = product (tree(:d-1)) + product (codes+777, mask=reson) * 11
    call cascade_compare (hash, tree, codes, reson, ok)
    if (ok) then
       k = n_cascades + 1
       if (k>size(cascade_set)) then
          write (msg_buffer, "(1x,A,1x,I6,1x,A)") &
               & "More than ", MAX_CASCADES, &
               & "cascades.  You may increase MAX_CASCADES in limits.f90"
          call msg_fatal
       end if
       cascade_hash(k) = hash
       call cascade_create (cascade_set(k), d)
       n_t_channel = count(tchan) - 1
       if (final)  n_t_channel = n_t_channel - 1
       if (resonant) then
          cascade_set(k)%multiplicity = 1
       else
          cascade_set(k)%multiplicity = &
               & cascade_set(i)%multiplicity &
               & + cascade_set(j)%multiplicity
       end if
       cascade_set(k)%active = .true.
       cascade_set(k)%depth = d
       cascade_set(k)%final = final
       cascade_set(k)%resonant = reson
       cascade_set(k)%n_resonances = n_resonances
       cascade_set(k)%t_channel = tchan
       cascade_set(k)%n_t_channel = n_t_channel
       cascade_set(k)%node%bincode = bc
       cascade_set(k)%node%code = code
       cascade_set(k)%node%index = k
       cascade_set(k)%node%mass = mass
       cascade_set(k)%node%resonant = resonant
       cascade_set(k)%node%needs_cut = needs_cut
       cascade_set(k)%node%t_channel = any(tchan)
       cascade_set(k)%node%final = final
       cascade_set(k)%node%decays = .true.
       cascade_set(k)%node%daughter1 => cascade_set(i)%node
       cascade_set(k)%node%daughter2 => cascade_set(j)%node
       cascade_set(k)%tree = tree
       cascade_set(k)%code = codes
       n_cascades = k
    end if
    deallocate(tree, codes, reson)
  end subroutine cascade_merge

  subroutine tree_sort (tree, code, resonant, t_channel)
    integer(kind=TC), dimension(:), intent(inout) :: tree
    integer, dimension(:), intent(inout) :: code
    logical, dimension(:), intent(inout) :: resonant, t_channel
    integer(kind=TC), dimension(size(tree)) :: tree_tmp
    integer, dimension(size(code)) :: code_tmp
    logical, dimension(size(resonant)) :: resonant_tmp, t_channel_tmp
    integer, dimension(1) :: pos
    integer :: i
    tree_tmp = tree
    code_tmp = code
    resonant_tmp = resonant
    t_channel_tmp = t_channel
    do i=size(tree),1,-1
       pos = maxloc(tree_tmp)
       tree(i) = tree_tmp(pos(1))
       code(i) = code_tmp(pos(1))
       resonant(i) = resonant_tmp(pos(1))
       t_channel(i) = t_channel_tmp(pos(1))
       tree_tmp(pos(1)) = 0
    end do
  end subroutine tree_sort

  subroutine cascade_check_decay &
       & (index, newcode, par, threshold_mass, &
       &  ok, resonant, needs_cut, newmass, mass_sum)
    integer, dimension(2), intent(in) :: index
    integer, intent(in) :: newcode
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    logical, intent(out) :: ok, resonant, needs_cut
    real(kind=default), intent(out) :: newmass, mass_sum
    integer, dimension(2) :: code
    logical, dimension(2) :: on_shell
    real(kind=default), dimension(2) :: mass, real_mass
    real(kind=default) :: width
    integer :: depth
    integer :: i
    do i=1,2
       code(i) = cascade_set(index(i))%node%code
       on_shell(i) = cascade_set(index(i))%node%resonant .or. &
            & .not. cascade_set(index(i))%node%decays
       mass(i) = cascade_set(index(i))%node%mass
       real_mass(i) = particle_mass(code(i), par)
       if (real_mass(i) < threshold_mass)  real_mass(i) = 0
    end do
    newmass = particle_mass (newcode, par)
    width = particle_width (newcode, par)
    if (newmass < threshold_mass)  newmass = 0
    mass_sum = sum(mass)
    if (mass_sum < threshold_mass)  mass_sum = 0
    ok = .false.
    resonant = .false.
    needs_cut = .false.
    depth = max(cascade_set(index(1))%depth, cascade_set(index(2))%depth)
    ! Resonance: on-shell decay
    if (all(on_shell) .and. mass_sum < newmass .and. width > 0) then
       ok = .true.
       resonant = .true.
    ! Resonance: off-shell decay
    else if (any(on_shell) .and. mass_sum < newmass) then
       if (single_off_shell_decays .and. sum(real_mass) >= newmass .and. &
            & width > 0) then
          ok = .true.
          resonant = .true.
       else
          ok = single_off_shell_branchings
       end if
    ! Resonance: double off-shell decay
    else if (mass_sum < newmass) then
       if (double_off_shell_decays .and. sum(real_mass) >= newmass .and. &
            & width > 0) then
          ok = .true.
          resonant = .true.
       else
          ok = double_off_shell_branchings
       end if
    ! Massless splitting
    else if (massless_splittings .and. depth <= splitting_depth &
         &   .and. all(on_shell) &
         &   .and. newmass == 0 .and. mass_sum == 0 .and. sum(real_mass) == 0 &
         &  ) then
       ok = .true.
       resonant = .true.
    ! Massless radiation
    else if (mass(1)==0.and.on_shell(1) .and. &
         &     mass(2)==newmass.and.real_mass(2)==newmass  .or. &
         &   mass(2)==0.and.on_shell(2) .and. &
         &     mass(1)==newmass.and.real_mass(1)==newmass) then
       ok = .true.
       resonant = .false.
       needs_cut = .true.
    ! Virtual two-particle production
    else if (all(on_shell)) then
       ok = .true.
       resonant = .false.
       newmass = max(mass_sum, newmass)
    ! Branchings with one daughter off-shell
    else if (any(on_shell)) then
       ok = single_off_shell_branchings
       resonant = .false.
    ! Everything else
    else
       ok = double_off_shell_branchings
       resonant = .false.
    end if
  end subroutine cascade_check_decay

  subroutine cascade_check_exchange &
       & (index, newcode, par, threshold_mass, &
       &  final, ok, resonant, needs_cut, newmass)
    integer, dimension(2), intent(in) :: index
    integer, intent(in) :: newcode
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: threshold_mass
    logical, intent(in) :: final
    logical, intent(out) :: ok, resonant, needs_cut
    real(kind=default), intent(out) :: newmass
    logical :: initial
    integer, dimension(2) :: code
    logical :: on_shell
    real(kind=default), dimension(2) :: mass, real_mass
    integer :: i
    initial = cascade_set(index(1))%node%initial
    on_shell = cascade_set(index(2))%node%resonant .or. &
         & .not. cascade_set(index(2))%node%decays
    do i=1,2
       code(i) = cascade_set(index(i))%node%code
       mass(i) = cascade_set(index(i))%node%mass
       real_mass(i) = particle_mass(code(i), par)
       if (real_mass(i) < threshold_mass)  real_mass(i) = 0
    end do
    resonant = .false.
    needs_cut = .false.
    newmass = particle_mass (newcode, par)
    if (newmass < threshold_mass)  newmass = 0
    ! s-channel graphs
    if (initial.eqv.final) then
       ok = .true.
    ! off-shell particle production (massive FSR)
    else if (.not.on_shell .and. mass(2)>real_mass(2)) then
       ok = massive_FSR
    ! Massless ISR
    else if (initial .and. on_shell .and. &
         &   mass(1)==newmass .and. mass(2)==0) then
       ok = .true.
       needs_cut = .true.
    ! Multiple massless ISR
    else if (cascade_set(index(1))%node%needs_cut .and. &
         &   on_shell .and. mass(2)==0) then
       ok = .true.
       needs_cut = .true.
    ! Massless exchange: this will be mapped
    else if (final .and. on_shell .and. &
         &   mass(1)==0 .and. mass(2)==newmass)  then
       ok = .true.
       resonant = .true.
       needs_cut = .true.
    ! Everything else will be just kept
    else
       ok = .true.
    end if
  end subroutine cascade_check_exchange

  subroutine cascade_compare (hash, tree, code, resonant, ok)
    integer, intent(in) :: hash
    integer(kind=TC), dimension(:), intent(in) :: tree
    integer, dimension(:), intent(in) :: code
    logical, dimension(:), intent(in) :: resonant
    logical, intent(out) :: ok
    integer :: k, depth, d
    ok = .true.
    depth = size(tree)
    d = depth - 2
    COMPARE_TREES: do k=1, n_cascades
       if (.not.cascade_set(k)%active) cycle COMPARE_TREES
       if (hash /= cascade_hash(k)) cycle COMPARE_TREES
       if (cascade_set(k)%depth==depth) then
          if (all (cascade_set(k)%resonant.eqv.resonant)) then
             if (all (cascade_set(k)%tree==tree) .and. &
               & all (cascade_set(k)%code==code)) then
                ok = .false.
                exit COMPARE_TREES
             else if (cascade_set(k)%final) then
                if ( all (cascade_set(k)%tree==tree) .and. &
                  &  all (cascade_set(k)%code==code .or. .not.resonant)) then
                   ok = .false.
                   exit COMPARE_TREES
                else if (cascade_set(k)%n_t_channel == 0 .and. &
                     &   all (cascade_set(k)%tree(:d) == tree(:d))) then
                   ok = .false.
                   exit COMPARE_TREES
                end if
             end if
          else if (cascade_set(k)%final) then
             if (   all (cascade_set(k)%tree==tree) .and. &
                  & all (cascade_set(k)%code==code)) then
                if (cascade_set(k)%n_resonances > count(resonant)) then
                   ok = .false.
                   exit COMPARE_TREES
                else if (cascade_set(k)%n_resonances < count(resonant)) then
                   cascade_set(k)%active = .false.
                   ok = .true.
                   exit COMPARE_TREES
                end if
             end if
          end if
       end if
    end do COMPARE_TREES
  end subroutine cascade_compare

  subroutine cascade_set_select (off_shell, peripheral, extra, s_channel)
    integer, intent(in), optional :: off_shell, peripheral, extra
    logical, intent(in), optional :: s_channel
    integer :: max_res, offset, off2, offt
    logical :: schan
    offset = 0;  if (present(off_shell)) offset = abs(off_shell)
    offt   = 0;  if (present(peripheral)) offt  = abs(peripheral)
    off2   = 0;  if (present(extra))      off2  = abs(extra)
    schan = .false.;  if (present(s_channel)) schan = s_channel
    max_res = (maxval (cascade_set%depth, mask=cascade_set%active) - 3)/2
    where (cascade_set%final .and. cascade_set%active .and. &
         & cascade_set%n_resonances < max_res - offset - off2 - 1) &
         &   cascade_set%active = .false.
    where (cascade_set%final .and. cascade_set%active .and. &
         & cascade_set%n_resonances < max_res - offset - 1) &
         &   cascade_set%deleted = .true.
    where (cascade_set%final .and. cascade_set%active .and. &
         & cascade_set%n_resonances == max_res - offset - 1 .and. &
         & cascade_set%n_t_channel < cascade_set%multiplicity - offt) &
         &   cascade_set%deleted = .true.
    where (cascade_set%final .and. cascade_set%active .and. &
         & cascade_set%n_resonances < max_res - offset) &
         &   cascade_set%subdominant = .true.
    if (.not.schan) &
         & where (cascade_set%final .and. cascade_set%active .and. &
         &        cascade_set%multiplicity == 1) &
         &   cascade_set%deleted = .true.
  end subroutine cascade_set_select
         
  subroutine cascade_set_validate
    integer :: i
    do i=1, n_cascades
       if (.not.cascade_set(i)%final) then
          cascade_set(i)%active = .false.
       else if (cascade_set(i)%active) then
          call activate (cascade_set(i)%node)
       end if
    end do
  contains
    recursive subroutine activate (node)
      type(node_t), pointer :: node
      cascade_set(node%index)%active = .true.
      if (node%decays) then
         call activate (node%daughter1)
         call activate (node%daughter2)
      end if
    end subroutine activate
  end subroutine cascade_set_validate
      
  subroutine cascade_set_groves
    integer :: max_resonances, max_t_channel, min_mult
    integer :: n_r, n_t, mult
    logical, dimension(:), allocatable :: mask_active
    logical, dimension(:), allocatable :: mask_mult, mask_t, mask_r
    n_groves = 0
    allocate(mask_active(n_cascades))
    allocate(mask_mult(n_cascades), mask_t(n_cascades), mask_r(n_cascades))
    where (cascade_set(:n_cascades)%active)
       mask_active = cascade_set(:n_cascades)%final
    elsewhere
       mask_active = .false.
    end where
    if (count(mask_active) > 0) then
       min_mult = minval (cascade_set(:n_cascades)%multiplicity, &
            &             mask=mask_active)
       SCAN_MULT: do mult = min_mult, cascade_set(n_cascades)%depth
          where (mask_active)  
             mask_mult = cascade_set(:n_cascades)%multiplicity == mult
          elsewhere
             mask_mult = .false.
          end where
          if (count(mask_mult)>0) then
             max_resonances = maxval &
                  & (cascade_set(:n_cascades)%n_resonances, mask=mask_mult)
             SCAN_RES: do n_r=max_resonances, 0, -1
                where (mask_mult)
                   mask_r = cascade_set(:n_cascades)%n_resonances == n_r
                elsewhere
                   mask_r = .false.
                end where
                if (count(mask_r)>0) then
                   max_t_channel = maxval &
                        & (cascade_set(:n_cascades)%n_t_channel, mask=mask_r)
                   SCAN_T: do n_t=max_t_channel, 0, -1
                      where (mask_r)
                         mask_t = cascade_set(:n_cascades)%n_t_channel == n_t
                      elsewhere
                         mask_t = .false.
                      end where
                      if (count(mask_t)>0) then
                         call cascade_groves_by_resonances (n_r, n_t, mask_t)
                      end if
                   end do SCAN_T
                end if
             end do SCAN_RES
          end if
       end do SCAN_MULT
    end if
    deallocate(mask_active, mask_mult, mask_t, mask_r)
  end subroutine cascade_set_groves

  subroutine cascade_groves_by_resonances (n_resonances, n_t_channel, mask)
    integer, intent(in) :: n_resonances, n_t_channel
    logical, dimension(:), intent(in) :: mask
    logical, dimension(size(mask)) :: current_mask
    integer, dimension(n_resonances, size(mask)) :: resonances
    integer, dimension(n_t_channel, size(mask)) :: t_channel
    integer, dimension(n_resonances) :: res
    integer, dimension(n_t_channel) :: tch
    integer :: i
    logical :: first
    resonances = UNDEFINED
    t_channel = UNDEFINED
    do i=1, size(mask)
       if (mask(i)) then
          call transfer_array (resonances(:,i), &
               & abs(cascade_set(i)%code), mask=cascade_set(i)%resonant)
          call transfer_array (t_channel(:,i), &
               & abs(cascade_set(i)%code), mask=cascade_set(i)%t_channel)
       end if
    end do
    current_mask = mask
    SET_GROVES: do
       if (count(current_mask) == 0) exit SET_GROVES
       first = .true.
       do i=1, size(mask)
          if (current_mask(i)) then
             if (first) then
                n_groves = n_groves + 1
                res = resonances(:,i)
                tch = t_channel(:,i)
                cascade_set(i)%grove = n_groves
                current_mask(i) = .false.
                first = .false.
             else if (all (resonances(:,i)==res) .and. &
                  &   all (t_channel(:,i)==tch)) then
                cascade_set(i)%grove = n_groves
                current_mask(i) = .false.
             end if
          end if
       end do
    end do SET_GROVES
  end subroutine cascade_groves_by_resonances

  subroutine transfer_array (array, code, mask)
    integer, dimension(:), intent(out) :: array
    integer, dimension(:), intent(in) :: code
    logical, dimension(:), intent(in), optional :: mask
    logical, dimension(size(mask)) :: current_mask
    integer :: i
    integer, dimension(1) :: pos
    if (present(mask)) then
       current_mask = mask
    else
       current_mask = .true.
    end if
    do i=1, size(array)
       pos = maxloc(code, mask=current_mask)
       array(i) = code(pos(1))
       current_mask(pos(1)) = .false.
    end do
  end subroutine transfer_array

  subroutine make_phase_space_file &
       & (unit, prefix, read_model_file, process_id, &
       &  code, par, n_channels, phs)
    integer, intent(in) :: unit
    character(len=*), intent(in) :: prefix, read_model_file, process_id
    integer, dimension(:,:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer, intent(out) :: n_channels
    type(phase_space_switches), intent(in) :: phs
    integer :: n_out, i1, i2, depth, max_off_shell
    real(kind=default) :: threshold_mass
    single_off_shell_decays = phs%single_off_shell_decays
    double_off_shell_decays = phs%double_off_shell_decays
    single_off_shell_branchings = phs%single_off_shell_branchings
    double_off_shell_branchings = phs%double_off_shell_branchings
    massive_FSR = phs%massive_FSR
    massless_splittings = phs%massless_splittings
    splitting_depth = phs%splitting_depth
    threshold_mass = phs%threshold_mass
    n_out = size(code,dim=1) - 2
    depth = 2*n_out + 1
    max_off_shell = 1
    if (phs%exchange_lines==0) then
       max_off_shell = 0
    else
       max_off_shell = 1
    end if
    max_off_shell = max_off_shell + abs(phs%off_shell_lines)
    max_off_shell = max_off_shell + abs(phs%extra_off_shell_lines)
    call msg_message (" Generating phase space channels for process " &
         &            //trim(process_id)// "...")
!    call start_comment (6)
    call setup_and_read_vertices (prefix, read_model_file)
    call cascade_set_init (code(3:,:), par, threshold_mass)
    call cascade_set_generate (par, threshold_mass, max_off_shell)
!    write(*,'(A)', advance='no') "+"
    call cascade_set_add &
         & (ibset(0,n_out), antiparticle_code(code(2,1)), par, &
         &  threshold_mass, initial=.true., index=i1)
    call cascade_set_add &
         & (ibset(0,n_out+1), antiparticle_code(code(1,1)), par, &
         &  threshold_mass, initial=.true., index=i2)
    call cascade_set_generate &
         & (par, threshold_mass, depth, &
         &  seed=i1, target=i2, off_shell=max_off_shell)
!    write(*,'(A)', advance='no') "+"
    call cascade_set_generate &
         & (par, threshold_mass, depth, &
         &  seed=i2, target=i1, off_shell=max_off_shell)
!    write(*,'(A)') "+"
    call cascade_set_select (off_shell=phs%off_shell_lines, &
         & peripheral=phs%exchange_lines, extra=phs%extra_off_shell_lines, &
         & s_channel=phs%s_channel_resonance)
    call cascade_set_validate
    call cascade_set_groves
    n_channels = count(cascade_set%final .and. cascade_set%active &
         &             .and. .not.cascade_set%deleted)
    if (n_channels==0) then
       call msg_message (" No valid phase space channels could be generated.")
       call msg_message (" You may need to increase, e.g., off_shell_lines in the input file")
       call msg_fatal (" Phase space configuration setup failed.")
    else
       write (msg_buffer, "(1x,A,1x,I5,1x,A)") &
            & "Phase space:", n_channels, "phase space channels generated."
       call msg_message
    end if
!    call cascade_set_write (6, final=.false.)
    call msg_message (repeat("-",72), unit=unit)
    call msg_message &
         & (" Automatically generated phase space configuration", unit=unit)
    call write_process (unit, process_id, code(:,1), 2, n_out)
    write (unit, '(A,1x,A)') "process", trim(process_id)
    call cascade_set_write (unit, final=.true., short=.true., sort=.true.)
    write (unit, *)
    call msg_message (" Dummy entry", unit=unit)
    write (unit, '(A,1x,A)') "process", trim(process_id)
    call cascade_set_clear
    call clear_vertices
  end subroutine make_phase_space_file


end module cascades
